//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002-2004
//	Sleepycat Software.  All rights reserved.
//
// $Id: dbxmlsh.cpp,v 1.7 2004/11/22 21:04:07 gmf Exp $
//

#include <fstream>
#include <string>

#include "dbxml/DbXml.hpp"
#include "Environment.hpp"
#include "DefaultShell.hpp"

using namespace DbXml;
using namespace std;

void usage(const string &progname, int exitCode)
{
	int pos = progname.rfind('/');
	if(pos == string::npos) {
		pos = progname.rfind('\\');
	}
	if(pos == string::npos) {
		pos = 0;
	}
	else {
		++pos;
	}

	cerr << "Usage: " << progname.substr(pos) << " [OPTIONS] path_to_dbenv (default_container)?" << endl << endl;
	cerr << "Options:" << endl;
	cerr << "\t-x\t\trun in secure mode, meaning there is no access to the local file system or the internet" << endl;
	cerr << "\t-t\t\trun in transaction mode - meaning transactions can be used, and are required for writes" << endl;
	cerr << "\t-s script\trun the script specified, in non-interactive mode" << endl;
	cerr << "\t-v\t\toutput more verbosely. Using the switch a second time increases verbosity" << endl;
	cerr << "\t-h\t\tprint this help" << endl;
	cerr << endl;
	exit(exitCode);
}

int main(int argc, char **argv)
{
	// Deal with command line arguments
	string path2DbEnv;
	u_int32_t envFlags = (DB_CREATE|DB_INIT_LOCK|DB_INIT_LOG|DB_INIT_MPOOL|DB_INIT_TXN);
	u_int32_t dbxmlFlags = DBXML_ALLOW_EXTERNAL_ACCESS;
	string theContainer;
	vector<string> scripts;
	int verbose = 0;
	bool transactionMode = false;
	for(int i=1; i<argc; ++i) {
		if(argv[i][0] == '-') {
			switch(argv[i][1]) {
			case 'h': {
				usage(argv[0], 0);
				break;
			}
			case 'x': {
				dbxmlFlags &= ~DBXML_ALLOW_EXTERNAL_ACCESS;
				break;
			}
			case 't': {
				transactionMode = true;
				break;
			}
			case 's': {
				if(++i == argc) {
					cerr << "No script specified" << endl;
					usage(argv[0], -1);
				}
				else {
					scripts.push_back(argv[i]);
				}
				break;
			}
			case 'v': {
				++verbose;
				break;
			}
			default: {
				cerr << "Unknown option: " << argv[i] << endl;
				usage(argv[0], -1);
				break;
			}
			}
		}
		else if(!path2DbEnv.length()) {
			path2DbEnv = argv[i];
		}
		else if(!theContainer.length()) {
			theContainer = argv[i];
		}
		else {
			cerr << "Too many arguments" << endl;
			usage(argv[0], -1);
		}
	}

	if(!path2DbEnv.length()) {
		cerr << "No path to DbEnv specified" << endl;
		usage(argv[0], -1);
	}

	// Turn on logging if extra verbose is specified
	if(verbose > 1) {
		setLogLevel(LEVEL_ALL, true);
		setLogCategory(CATEGORY_ALL, true);
		setLogCategory(CATEGORY_NODESTORE, verbose > 2);
	}

	try {
		// Create a database
		DbEnv dbenv(0);
		dbenv.set_error_stream(&cerr);
		dbenv.set_cachesize(0, 64 * 1024, 1);
		dbenv.open(path2DbEnv.c_str(), envFlags, 0);
		XmlManager db(&dbenv, dbxmlFlags);

		// Create the environment
		Environment env(db);
		env.transactions() = transactionMode;

		// Open the default container, if we have one
		if(theContainer.length()) {
			if(env.transactions()) {
				XmlTransaction myTxn = db.createTransaction();
				env.container() = new XmlContainer(db.openContainer(myTxn, theContainer));
				myTxn.commit();
			}
			else {
				env.container() = new XmlContainer(db.openContainer(theContainer));
			}
		}

		// Create the Shell object
		DefaultShell shell;

		// Run the scripts, if specified
		if(!scripts.empty()) {

			env.interactive() = false;
			env.verbose() = verbose != 0;

			for(vector<string>::iterator i = scripts.begin();
			    i != scripts.end() && !env.quit(); ++i) {
				ifstream scriptFile(i->c_str(), ios::in);
				if(!scriptFile) {
					cerr << "Cannot open file: " << *i << endl;
				}
				else {
					env.streamName() = *i;
					env.lineNo() = 0;
					shell.mainLoop(scriptFile, env);
					scriptFile.close();
				}
			}
		}

		// Perform the queries
		if(!env.quit()) {
			env.interactive() = true;
			env.verbose() = true;
			env.streamName() = "stdin";
			env.lineNo() = 0;
			shell.mainLoop(cin, env);
		}
	}
	catch(exception &e) {
		cerr << "Error at lowest level: " << e.what() << endl;
		return 1;
	}
	catch(...) {
		cerr << "Error at lowest level" << endl;
		return 1;
	}

	return 0;
}
